# Perforce source for convert extension.
#
# Copyright 2009, Frank Kingswood <frank@kingswood-consulting.co.uk>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import marshal
import re

from mercurial.i18n import _
from mercurial import (
    error,
    util,
)
from mercurial.utils import (
    dateutil,
    procutil,
    stringutil,
)

from . import common


def loaditer(f):
    """Yield the dictionary objects generated by p4"""
    try:
        while True:
            d = marshal.load(f)
            if not d:
                break
            yield d
    except EOFError:
        pass


def decodefilename(filename):
    """Perforce escapes special characters @, #, *, or %
    with %40, %23, %2A, or %25 respectively

    >>> decodefilename(b'portable-net45%252Bnetcore45%252Bwp8%252BMonoAndroid')
    'portable-net45%2Bnetcore45%2Bwp8%2BMonoAndroid'
    >>> decodefilename(b'//Depot/Directory/%2525/%2523/%23%40.%2A')
    '//Depot/Directory/%25/%23/#@.*'
    """
    replacements = [
        (b'%2A', b'*'),
        (b'%23', b'#'),
        (b'%40', b'@'),
        (b'%25', b'%'),
    ]
    for k, v in replacements:
        filename = filename.replace(k, v)
    return filename


class p4_source(common.converter_source):
    def __init__(self, ui, repotype, path, revs=None):
        # avoid import cycle
        from . import convcmd

        super().__init__(ui, repotype, path, revs=revs)

        if b"/" in path and not path.startswith(b'//'):
            raise common.NoRepo(
                _(b'%s does not look like a P4 repository') % path
            )

        common.checktool(b'p4', abort=False)

        self.revmap = {}
        self.encoding = self.ui.config(
            b'convert', b'p4.encoding', convcmd.orig_encoding
        )
        self.re_type = re.compile(
            br"([a-z]+)?(text|binary|symlink|apple|resource|unicode|utf\d+)"
            br"(\+\w+)?$"
        )
        self.re_keywords = re.compile(
            br"\$(Id|Header|Date|DateTime|Change|File|Revision|Author)"
            br":[^$\n]*\$"
        )
        self.re_keywords_old = re.compile(br"\$(Id|Header):[^$\n]*\$")

        if revs and len(revs) > 1:
            raise error.Abort(
                _(
                    b"p4 source does not support specifying "
                    b"multiple revisions"
                )
            )

    def setrevmap(self, revmap):
        """Sets the parsed revmap dictionary.

        Revmap stores mappings from a source revision to a target revision.
        It is set in convertcmd.convert and provided by the user as a file
        on the commandline.

        Revisions in the map are considered beeing present in the
        repository and ignored during _parse(). This allows for incremental
        imports if a revmap is provided.
        """
        self.revmap = revmap

    def _parse_view(self, path):
        """Read changes affecting the path"""
        cmd = b'p4 -G changes -s submitted %s' % procutil.shellquote(path)
        stdout = procutil.popen(cmd, mode=b'rb')
        p4changes = {}
        for d in loaditer(stdout):
            c = d.get(b"change", None)
            if c:
                p4changes[c] = True
        return p4changes

    def _parse(self, ui, path):
        """Prepare list of P4 filenames and revisions to import"""
        p4changes = {}
        changeset = {}
        files_map = {}
        copies_map = {}
        localname = {}
        depotname = {}
        heads = []

        ui.status(_(b'reading p4 views\n'))

        # read client spec or view
        if b"/" in path:
            p4changes.update(self._parse_view(path))
            if path.startswith(b"//") and path.endswith(b"/..."):
                views = {path[:-3]: b""}
            else:
                views = {b"//": b""}
        else:
            cmd = b'p4 -G client -o %s' % procutil.shellquote(path)
            clientspec = marshal.load(procutil.popen(cmd, mode=b'rb'))

            views = {}
            for client in clientspec:
                if client.startswith(b"View"):
                    sview, cview = clientspec[client].split()
                    p4changes.update(self._parse_view(sview))
                    if sview.endswith(b"...") and cview.endswith(b"..."):
                        sview = sview[:-3]
                        cview = cview[:-3]
                    cview = cview[2:]
                    cview = cview[cview.find(b"/") + 1 :]
                    views[sview] = cview

        # list of changes that affect our source files
        p4changes = sorted(p4changes.keys(), key=int)

        # list with depot pathnames, longest first
        vieworder = sorted(views.keys(), key=len, reverse=True)

        # handle revision limiting
        startrev = self.ui.config(b'convert', b'p4.startrev')

        # now read the full changelists to get the list of file revisions
        ui.status(_(b'collecting p4 changelists\n'))
        lastid = None
        for change in p4changes:
            if startrev and int(change) < int(startrev):
                continue
            if self.revs and int(change) > int(self.revs[0]):
                continue
            if change in self.revmap:
                # Ignore already present revisions, but set the parent pointer.
                lastid = change
                continue

            if lastid:
                parents = [lastid]
            else:
                parents = []

            d = self._fetch_revision(change)
            c = self._construct_commit(d, parents)

            descarr = c.desc.splitlines(True)
            if len(descarr) > 0:
                shortdesc = descarr[0].rstrip(b'\r\n')
            else:
                shortdesc = b'**empty changelist description**'

            t = b'%s %s' % (c.rev, shortdesc)
            ui.status(stringutil.ellipsis(t, 80) + b'\n')

            files = []
            copies = {}
            copiedfiles = []
            i = 0
            while (b"depotFile%d" % i) in d and (b"rev%d" % i) in d:
                oldname = d[b"depotFile%d" % i]
                filename = None
                for v in vieworder:
                    if oldname.lower().startswith(v.lower()):
                        filename = decodefilename(views[v] + oldname[len(v) :])
                        break
                if filename:
                    files.append((filename, d[b"rev%d" % i]))
                    depotname[filename] = oldname
                    if d.get(b"action%d" % i) == b"move/add":
                        copiedfiles.append(filename)
                    localname[oldname] = filename
                i += 1

            # Collect information about copied files
            for filename in copiedfiles:
                oldname = depotname[filename]

                flcmd = b'p4 -G filelog %s' % procutil.shellquote(oldname)
                flstdout = procutil.popen(flcmd, mode=b'rb')

                copiedfilename = None
                for d in loaditer(flstdout):
                    copiedoldname = None

                    i = 0
                    while (b"change%d" % i) in d:
                        if (
                            d[b"change%d" % i] == change
                            and d[b"action%d" % i] == b"move/add"
                        ):
                            j = 0
                            while (b"file%d,%d" % (i, j)) in d:
                                if d[b"how%d,%d" % (i, j)] == b"moved from":
                                    copiedoldname = d[b"file%d,%d" % (i, j)]
                                    break
                                j += 1
                        i += 1

                    if copiedoldname and copiedoldname in localname:
                        copiedfilename = localname[copiedoldname]
                        break

                if copiedfilename:
                    copies[filename] = copiedfilename
                else:
                    ui.warn(
                        _(b"cannot find source for copied file: %s@%s\n")
                        % (filename, change)
                    )

            changeset[change] = c
            files_map[change] = files
            copies_map[change] = copies
            lastid = change

        if lastid and len(changeset) > 0:
            heads = [lastid]

        return {
            b'changeset': changeset,
            b'files': files_map,
            b'copies': copies_map,
            b'heads': heads,
            b'depotname': depotname,
        }

    @util.propertycache
    def _parse_once(self):
        return self._parse(self.ui, self.path)

    @util.propertycache
    def copies(self):
        return self._parse_once[b'copies']

    @util.propertycache
    def files(self):
        return self._parse_once[b'files']

    @util.propertycache
    def changeset(self):
        return self._parse_once[b'changeset']

    @util.propertycache
    def heads(self):
        return self._parse_once[b'heads']

    @util.propertycache
    def depotname(self):
        return self._parse_once[b'depotname']

    def getheads(self):
        return self.heads

    def getfile(self, name, rev):
        cmd = b'p4 -G print %s' % procutil.shellquote(
            b"%s#%s" % (self.depotname[name], rev)
        )

        lasterror = None
        while True:
            stdout = procutil.popen(cmd, mode=b'rb')

            mode = None
            contents = []
            keywords = None

            for d in loaditer(stdout):
                code = d[b"code"]
                data = d.get(b"data")

                if code == b"error":
                    # if this is the first time error happened
                    # re-attempt getting the file
                    if not lasterror:
                        lasterror = IOError(d[b"generic"], data)
                        # this will exit inner-most for-loop
                        break
                    else:
                        raise lasterror

                elif code == b"stat":
                    action = d.get(b"action")
                    if action in [b"purge", b"delete", b"move/delete"]:
                        return None, None
                    p4type = self.re_type.match(d[b"type"])
                    if p4type:
                        mode = b""
                        flags = (p4type.group(1) or b"") + (
                            p4type.group(3) or b""
                        )
                        if b"x" in flags:
                            mode = b"x"
                        if p4type.group(2) == b"symlink":
                            mode = b"l"
                        if b"ko" in flags:
                            keywords = self.re_keywords_old
                        elif b"k" in flags:
                            keywords = self.re_keywords

                elif code == b"text" or code == b"binary":
                    contents.append(data)

                lasterror = None

            if not lasterror:
                break

        if mode is None:
            return None, None

        contents = b''.join(contents)

        if keywords:
            contents = keywords.sub(b"$\\1$", contents)
        if mode == b"l" and contents.endswith(b"\n"):
            contents = contents[:-1]

        return contents, mode

    def getchanges(self, rev, full):
        if full:
            raise error.Abort(_(b"convert from p4 does not support --full"))
        return self.files[rev], self.copies[rev], set()

    def _construct_commit(self, obj, parents=None):
        """
        Constructs a common.commit object from an unmarshalled
        `p4 describe` output
        """
        desc = self.recode(obj.get(b"desc", b""))
        date = (int(obj[b"time"]), 0)  # timezone not set
        if parents is None:
            parents = []

        return common.commit(
            author=self.recode(obj[b"user"]),
            date=dateutil.datestr(date, b'%Y-%m-%d %H:%M:%S %1%2'),
            parents=parents,
            desc=desc,
            branch=None,
            rev=obj[b'change'],
            extra={b"p4": obj[b'change'], b"convert_revision": obj[b'change']},
        )

    def _fetch_revision(self, rev):
        """Return an output of `p4 describe` including author, commit date as
        a dictionary."""
        cmd = b"p4 -G describe -s %s" % rev
        stdout = procutil.popen(cmd, mode=b'rb')
        return marshal.load(stdout)

    def getcommit(self, rev):
        if rev in self.changeset:
            return self.changeset[rev]
        elif rev in self.revmap:
            d = self._fetch_revision(rev)
            return self._construct_commit(d, parents=None)
        raise error.Abort(
            _(b"cannot find %s in the revmap or parsed changesets") % rev
        )

    def gettags(self):
        return {}

    def getchangedfiles(self, rev, i):
        return sorted([x[0] for x in self.files[rev]])



# deltas.py - utilities to parse and combine deltas
#
# Copyright 2023 Laurent Bulteau <laurent.bulteau@univ-eiffel.fr>
# Copyright 2023 Octobus <contact@octobus.net>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
"""This module gather function to parse and combine deltas.

The `mpatch` module also contains function dealing with similar structure.
However while `mpatch` focus on efficient Python code, this module focus on
having Python version easy to read and clear to understand.
"""

from __future__ import annotations

import struct
import typing

from typing import (
    Iterator,
    Sequence,
    Tuple,
)

from ..thirdparty import attr

if typing.TYPE_CHECKING:
    import attr


@attr.s(slots=True, frozen=True, repr=False)
class DeltaPiece:
    """represent a piece of delta, replacing one section with another

    Used for debug, not intended to be performant.
    """

    start = attr.ib(type=int)
    old_end = attr.ib(type=int)
    size = attr.ib(type=int)
    new_end = attr.ib(
        type=int,
        default=attr.Factory(
            lambda s: s.start + s.size,
            takes_self=True,
        ),
    )
    storage_size = attr.ib(
        type=int,
        default=attr.Factory(
            lambda s: 12 + s.size,
            takes_self=True,
        ),
    )
    offset = attr.ib(
        type=int,
        default=attr.Factory(
            lambda s: s.size - (s.old_end - s.start),
            takes_self=True,
        ),
    )

    def __le__(self, other):
        return self.start <= other.start

    def __repr__(self) -> str:
        return f"DeltaPiece({self.start},{self.old_end},{self.size})"

    def offsetted(self, offset):
        """return a new version of this DeltaPiece moved by <offset> bytes


        >>> DeltaPiece(0, 10, 15).offsetted(7)
        DeltaPiece(7,17,15)
        >>> DeltaPiece(20, 30, 12).offsetted(-5)
        DeltaPiece(15,25,12)
        """
        if offset == 0:
            return self
        assert 0 <= self.start + offset, (self.start, offset)
        return DeltaPiece(self.start + offset, self.old_end + offset, self.size)

    def truncate(self, amount):
        """return a new version of this DeltaPiece, removing the <amount> first bytes change to the base

        >>> DeltaPiece(10, 20, 3).truncate(5)
        DeltaPiece(15,20,0)
        >>> DeltaPiece(10, 20, 5).truncate(5)
        DeltaPiece(15,20,0)
        >>> DeltaPiece(10, 20, 10).truncate(5)
        DeltaPiece(15,20,5)
        >>> DeltaPiece(10, 20, 15).truncate(5)
        DeltaPiece(15,20,10)
        """
        assert amount < (self.old_end - self.start)
        return DeltaPiece(
            self.start + amount,
            self.old_end,
            self.size - min(amount, self.size),
        )


@attr.s(slots=True, init=False, repr=False)
class Delta:
    """Represent a series of change against a base."""

    _patches = attr.ib(type=list[DeltaPiece])
    _storage_size = attr.ib(type=int)

    def __init__(
        self,
        patches: list[DeltaPiece],
    ) -> None:
        self._storage_size = None
        self._patches = patches

    @classmethod
    def from_bytes(cls, binary: bytes):
        """build a Delta from its binary representation"""
        pos = 0
        patches = []
        while pos < len(binary):
            h = struct.unpack(b">lll", binary[pos : pos + 12])
            p_start, p_end, p_size = h
            assert p_size >= 0
            pos += 12
            pos += p_size
            patches.append(DeltaPiece(p_start, p_end, p_size))

        return cls(patches)

    def __repr__(self) -> str:
        self.storage_size
        patches = [(p.start, p.old_end, p.size) for p in self._patches]
        return f"Delta(size={self.storage_size}, {patches})"

    @property
    def storage_size(self):
        """How many bytes are needed to represent this delta in binary"""
        if self._storage_size is None:
            self._storage_size = sum(p.storage_size for p in self._patches)
        return self._storage_size

    def combine(self, other: Delta) -> Delta:
        """combined a delta with another one

        The `self` object should be the earlier delta in the chain, the `other`
        object applying over it.
        This produce a new delta that would result in `other` content when
        applied on `self` base.

        input:
        A --[self]--> B --[other]-->C
        output:
        A --[combine]--> C

        Resulting delta is always "correct", but there is some case where is it not "minimal"

        For example, if a line is reverted to a previous content (A → B → A),
        folding the two last patch would result in a patch requesting an
        unecessary replacement of the line with "A".


        ###### test case without offset

        ### Check that disjoint patches do not get merged
        >>> low = Delta([DeltaPiece(0, 10, 10)])
        >>> high = Delta([DeltaPiece(20, 30, 10)])
        >>> low.storage_size
        22
        >>> high.storage_size
        22
        >>> low.combine(high)
        Delta(size=44, [(0, 10, 10), (20, 30, 10)])

        ### Check that consecutive patches get merged
        >>> low = Delta([DeltaPiece(0, 10, 10)])
        >>> high = Delta([DeltaPiece(10, 20, 10)])
        >>> low.storage_size
        22
        >>> high.storage_size
        22
        >>> low.combine(high)
        Delta(size=32, [(0, 20, 20)])

        ### Check that overlapping patches get merged
        >>> low = Delta([DeltaPiece(0, 10, 10)])
        >>> high = Delta([DeltaPiece(5, 15, 10)])
        >>> low.storage_size
        22
        >>> high.storage_size
        22
        >>> low.combine(high)
        Delta(size=27, [(0, 15, 15)])

        ### Check that a chain overlapping patches get merged
        >>> low = Delta([DeltaPiece(0, 10, 10), DeltaPiece(15, 25, 10), DeltaPiece(30, 40, 10)])
        >>> high = Delta([DeltaPiece(8, 18, 10), DeltaPiece(23, 33, 10)])
        >>> low.storage_size
        66
        >>> high.storage_size
        44
        >>> low.combine(high)
        Delta(size=52, [(0, 40, 40)])

        ### Check that a chain with patch being a super set than other get merged fine
        >>> low = Delta([DeltaPiece(0, 30, 30), DeltaPiece(50, 60, 10)])
        >>> high = Delta([DeltaPiece(10, 20, 10), DeltaPiece(40, 70, 30)])
        >>> low.storage_size
        64
        >>> high.storage_size
        64
        >>> low.combine(high)
        Delta(size=84, [(0, 30, 30), (40, 70, 30)])

        ###### test case without offset

        ### Check that disjoint patches do not get merged
        >>> low = Delta([DeltaPiece(0, 10, 5)])
        >>> high = Delta([DeltaPiece(10, 15, 10)])
        >>> low.storage_size
        17
        >>> high.storage_size
        22
        >>> low.combine(high)
        Delta(size=39, [(0, 10, 5), (15, 20, 10)])

        ### Check that consecutive patches get merged
        >>> low = Delta([DeltaPiece(0, 10, 15)])
        >>> high = Delta([DeltaPiece(15, 20, 7)])
        >>> low.storage_size
        27
        >>> high.storage_size
        19
        >>> low.combine(high)
        Delta(size=34, [(0, 15, 22)])

        ### Check that overlapping patches get merged
        >>> low = Delta([DeltaPiece(10, 20, 20)])
        >>> high = Delta([DeltaPiece(25, 35, 3)])
        >>> low.storage_size
        32
        >>> high.storage_size
        15
        >>> low.combine(high)
        Delta(size=30, [(10, 25, 18)])

        ### Check that a chain overlapping patches get merged (patches around that)
        >>> low = Delta([
        ...     DeltaPiece(0, 0, 5),
        ...     DeltaPiece(5, 10, 15),
        ...     DeltaPiece(15, 20, 0),
        ...     DeltaPiece(30, 35, 15),
        ...     DeltaPiece(40, 80, 10),
        ... ])
        >>> high = Delta([
        ...     DeltaPiece(18, 45, 5),
        ...     DeltaPiece(50, 65, 30),
        ...     DeltaPiece(75, 75, 5)
        ... ])
        >>> low.storage_size
        105
        >>> high.storage_size
        76
        >>> low.combine(high)
        Delta(size=101, [(0, 0, 5), (5, 82, 55), (85, 85, 5)])

        ### Check that a chain with patch being a super set than other get merged fine
        >>> low = Delta([
        ...     DeltaPiece(5, 10, 15),    # shift things around at the start (+10)
        ...     DeltaPiece(20, 50, 10),   # Bigger range than the higher one (-20)
        ...     DeltaPiece(80, 90, 40),   # smaller range than the higher one (+30)
        ... ])
        >>> high = Delta([
        ...     DeltaPiece(33, 38, 65),
        ...     DeltaPiece(65, 120, 10),
        ...     DeltaPiece(150, 200, 66),
        ... ])
        >>> low.storage_size
        101
        >>> high.storage_size
        177
        >>> low.combine(high)
        Delta(size=209, [(5, 10, 15), (20, 50, 70), (75, 100, 10), (130, 180, 66)])

        #### complexe case flagged by the tests
        >>> high = Delta([
        ...     DeltaPiece(0, 1652, 1652),
        ...     DeltaPiece(9869, 9902, 33),
        ...     DeltaPiece(29669, 29702, 33),
        ...     DeltaPiece(49469, 49502, 33),
        ...     DeltaPiece(59369, 59402, 33),
        ...     DeltaPiece(69269, 69302, 33),
        ...     DeltaPiece(69335, 69368, 33),
        ...     DeltaPiece(79169, 79202, 33),
        ...     DeltaPiece(98969, 99002, 33),
        ...     DeltaPiece(108869, 108902, 33),
        ...     DeltaPiece(118769, 118802, 33),
        ...     DeltaPiece(138569, 138602, 33),
        ...     DeltaPiece(158369, 158402, 33),
        ...     DeltaPiece(168269, 168302, 33),
        ...     DeltaPiece(178169, 178202, 33),
        ...     DeltaPiece(178235, 178268, 33),
        ...     DeltaPiece(188069, 188102, 33),
        ...     DeltaPiece(207869, 207902, 33),
        ...     DeltaPiece(217769, 217802, 33),
        ...     DeltaPiece(227669, 227702, 33),
        ...     DeltaPiece(247469, 247502, 33),
        ...     DeltaPiece(267269, 267302, 33),
        ...     DeltaPiece(277169, 277202, 33),
        ...     DeltaPiece(287069, 287102, 33),
        ...     DeltaPiece(287135, 287168, 33),
        ...     DeltaPiece(296969, 297002, 33),
        ...     DeltaPiece(316769, 316802, 33),
        ...     DeltaPiece(326669, 326702, 33),
        ...     DeltaPiece(336569, 336602, 33)
        ... ])
        >>> low = Delta([
        ...     DeltaPiece(0, 1651, 1652),
        ...     DeltaPiece(9868, 9901, 33),
        ...     DeltaPiece(19768, 19801, 33),
        ...     DeltaPiece(29668, 29701, 33),
        ...     DeltaPiece(39568, 39601, 33),
        ...     DeltaPiece(49468, 49501, 33),
        ...     DeltaPiece(59368, 59401, 33),
        ...     DeltaPiece(69268, 69301, 33),
        ...     DeltaPiece(79168, 79201, 33),
        ...     DeltaPiece(89068, 89101, 33),
        ...     DeltaPiece(98968, 99001, 33),
        ...     DeltaPiece(108868, 108901, 33),
        ...     DeltaPiece(118768, 118801, 33),
        ...     DeltaPiece(128668, 128701, 33),
        ...     DeltaPiece(138568, 138601, 33),
        ...     DeltaPiece(148468, 148501, 33),
        ...     DeltaPiece(158368, 158401, 33),
        ...     DeltaPiece(168268, 168301, 33),
        ...     DeltaPiece(178168, 178201, 33),
        ...     DeltaPiece(188068, 188101, 33),
        ...     DeltaPiece(197968, 198001, 33),
        ...     DeltaPiece(207868, 207901, 33),
        ...     DeltaPiece(217768, 217801, 33),
        ...     DeltaPiece(227668, 227701, 33),
        ...     DeltaPiece(237568, 237601, 33),
        ...     DeltaPiece(247468, 247501, 33),
        ...     DeltaPiece(257368, 257401, 33),
        ...     DeltaPiece(267268, 267301, 33),
        ...     DeltaPiece(277168, 277201, 33),
        ...     DeltaPiece(287068, 287101, 33),
        ...     DeltaPiece(296968, 297001, 33),
        ...     DeltaPiece(306868, 306901, 33),
        ...     DeltaPiece(316768, 316801, 33),
        ...     DeltaPiece(326668, 326701, 33),
        ...     DeltaPiece(336568, 336601, 33)
        ... ])
        >>> low.storage_size
        3194
        >>> high.storage_size
        2924
        >>> low.combine(high)
        Delta(size=3329, [(0, 1651, 1652), (9868, 9901, 33), (19768, 19801, 33), (29668, 29701, 33), (39568, 39601, 33), (49468, 49501, 33), (59368, 59401, 33), (69268, 69301, 33), (69334, 69367, 33), (79168, 79201, 33), (89068, 89101, 33), (98968, 99001, 33), (108868, 108901, 33), (118768, 118801, 33), (128668, 128701, 33), (138568, 138601, 33), (148468, 148501, 33), (158368, 158401, 33), (168268, 168301, 33), (178168, 178201, 33), (178234, 178267, 33), (188068, 188101, 33), (197968, 198001, 33), (207868, 207901, 33), (217768, 217801, 33), (227668, 227701, 33), (237568, 237601, 33), (247468, 247501, 33), (257368, 257401, 33), (267268, 267301, 33), (277168, 277201, 33), (287068, 287101, 33), (287134, 287167, 33), (296968, 297001, 33), (306868, 306901, 33), (316768, 316801, 33), (326668, 326701, 33), (336568, 336601, 33)])
        """
        new_patches = []

        # == delta naming ==
        #
        # the lower delta is called "bottom",
        # the higher delta is called "top",
        # the final result if called "folded"
        #
        # == content name ==
        #
        # the unknown content modified by "lower" and "folded" is called "base" or "A"
        # the result of apply "bottom" on A is B
        # the result of apply "top" on B is C
        #
        #   A ─┬─[bottom]─> B ─[top]─┬─> C
        #      └────────[folded]─────┘

        iter_bottom = iter(self._patches)
        iter_top = iter(other._patches)
        current_bottom = next(iter_bottom, None)
        current_top = next(iter_top, None)

        # The amount of bytes added or removed by content from "bottom" so far
        #
        # The reverse of this value need to be apply to "top" patches "start"
        # when considering where they will apply in "base" (and, therefor, how
        # they need to be represented in "folded")
        #
        # When the "low" patch:
        # - add N extra bytes of content compared the "base", it goes up by N
        # - remove N bytes of content compared the "base", it goes down by N
        #
        # For a patch "b" from bottom:
        #   - start of the resulting area in B is b.start + offset_BA
        #   - end of the resulting area in B is b.start + b.size + offset_BA
        #
        # For a patch "t" from top:
        #   - start of the replaced area in A is t.start - offset_BA
        #   - end of the replaced area in A is t.old_end - offset_BA
        offset_BA = 0

        while current_bottom is not None or current_top is not None:
            if current_top is None:
                # There is no "top" patches anymore
                #
                # Let us consume the remaining "bottom" ones.
                # We will exit the loop in the next iteration.
                new_patches.append(current_bottom)
                new_patches.extend(iter_bottom)
                current_bottom = None
            elif current_bottom is None:
                # There is no "bottom" patches anymore
                #
                # Let us consume the remaining "top" ones.
                # We will exit the loop in the next iteration.
                new_patches.append(current_top.offsetted(-offset_BA))
                for p in iter_top:
                    new_patches.append(p.offsetted(-offset_BA))
                current_top = None
            elif (current_bottom.start + current_bottom.size) < (
                current_top.start - offset_BA
            ):
                # The next patch in "bottom" alter an area of "base" that is
                # untouched by the next patch in "top"
                #
                # we can simply copy that "bottom" patch in "folded"
                new_patches.append(current_bottom)
                offset_BA += current_bottom.offset
                current_bottom = next(iter_bottom, None)
            elif (current_top.old_end - offset_BA) < current_bottom.start:
                # The next patch in "top" alter an aread of base untouched by
                # the next patch in "top"
                #
                # we can copy over that "top" patch after adjusting it location
                p = current_top.offsetted(-offset_BA)
                new_patches.append(p)
                current_top = next(iter_top, None)
            else:
                # There is some overlap between the next patch in "bottom" and
                # the next patch in "top" and they need to be merged.

                # start with a zero size area from the first patch
                start_A = min(
                    current_bottom.start, current_top.start - offset_BA
                )
                end_A = start_A

                # check content offset from each layer
                #
                # We don't consume the initial patch yet, as consuming them is
                # no different than consuming the subsequent one.
                local_offset_from_bottom = 0
                local_offset_from_top = 0

                # We look for successive patches in either list the overlap
                # with the current one, and merge them (extend end points and
                # adapt offsets)
                while not (current_bottom is None and current_top is None):
                    if current_top is None:
                        # We no longer have patch from "bottom", merge everything we can from "top"
                        while current_bottom is not None and (
                            current_bottom.start <= end_A
                        ):
                            if current_bottom.old_end > end_A:
                                end_A = current_bottom.old_end
                            local_offset_from_bottom += current_bottom.offset
                            current_bottom = next(iter_bottom, None)
                        break
                    elif current_bottom is None:
                        local_BA = offset_BA + local_offset_from_bottom
                        # We no longer have patch from "bottom", merge everything we can from "top"
                        while (
                            current_top is not None
                            and current_top.start - local_BA <= end_A
                        ):
                            if current_top.old_end - local_BA > end_A:
                                end_A = current_top.old_end - local_BA
                            local_offset_from_top += current_top.offset
                            current_top = next(iter_top, None)
                        break

                    # compute where each current patch starts
                    start_bottom = current_bottom.start
                    start_top = (
                        current_top.start - offset_BA - local_offset_from_bottom
                    )

                    if start_bottom > end_A and start_top > end_A:
                        # No remaining overlap, go to conclusion of this merge
                        break

                    # check if any patch is ahead of the other and by how much.
                    #
                    # If one is ahead of the other, we process it first. Either
                    # by consuming it entirely if there is no overlap with the
                    # next one, or by processing the part ahead.
                    #
                    # If the two patch start at the same level, consume the one contained in the other
                    distance = start_bottom - start_top
                    if distance < 0:
                        distance *= -1
                        size_A = current_bottom.old_end - current_bottom.start
                        if size_A <= distance:
                            # "top" does not overlap with the current patch, merge wholly
                            if current_bottom.old_end > end_A:
                                end_A = current_bottom.old_end
                            local_offset_from_bottom += current_bottom.offset
                            current_bottom = next(iter_bottom, None)
                        else:
                            # truncate the start to align it with the other one
                            end_A += distance
                            local_offset_from_bottom += min(
                                0, current_bottom.size - distance
                            )
                            current_bottom = current_bottom.truncate(distance)
                    elif distance > 0:
                        size_B = current_top.old_end - current_top.start
                        if size_B <= distance:
                            # "bottom" patch does not overlap with the current patch, merge wholly
                            end = (
                                current_top.old_end
                                - offset_BA
                                - local_offset_from_bottom
                            )
                            if end > end_A:
                                end_A = end
                            local_offset_from_top += current_top.offset
                            current_top = next(iter_top, None)
                        else:
                            # truncate the current "top" patch top align it with the "bottom" one
                            end_A += distance
                            local_offset_from_top += min(
                                0, current_top.size - distance
                            )
                            current_top = current_top.truncate(distance)
                    else:
                        assert start_bottom == start_top
                        # Check if the "top" patch only replace "bottom" content
                        size_B = current_top.old_end - current_top.start
                        if size_B <= current_bottom.size:
                            # "top" only replace bottom content, adjust the
                            # final patch size and rely on "bottom" to update
                            # the `end_A` value.
                            local_offset_from_top += current_top.offset
                            current_top = next(iter_top, None)
                        else:
                            # "top" replace more that the current "bottom",
                            # truncate the part associated with that "bottom"
                            # patch and consume it.
                            local_offset_from_top += min(
                                0, current_top.size - current_bottom.size
                            )
                            current_top = current_top.truncate(
                                current_bottom.size
                            )

                            # consume the "bottom" patch
                            end_A += (
                                current_bottom.old_end - current_bottom.start
                            )
                            local_offset_from_bottom += current_bottom.offset
                            current_bottom = next(iter_bottom, None)
                size = (
                    end_A
                    - start_A
                    + local_offset_from_top
                    + local_offset_from_bottom
                )
                new_patches.append(DeltaPiece(start_A, end_A, size))
                offset_BA += local_offset_from_bottom
        return Delta(new_patches)


def estimate_combined_deltas_size(deltas: Sequence[bytes]) -> int:
    """estimate an upper bound of the size of combiner delta_low with delta_high"""
    # filter empty deltas
    deltas = [d for d in deltas if d]
    if not deltas:
        return 0
    high = Delta.from_bytes(deltas[-1])
    for low_data in deltas[-2::-1]:
        if not low_data:
            continue
        low = Delta.from_bytes(low_data)
        high = low.combine(high)
    return high.storage_size


def optimize_base(
    delta: bytes,
    next_deltas: Iterator[Tuple[int, bytes]],
    max_size: int,
) -> int | None:
    """Try to find a better base than the existing delta

    This function starts from <delta>, fold with other candidates from
    <next_deltas> until the estimated size gets over <max_size>.
    """
    if not delta:
        # if the delta is empty, we won't do better unless the rest of the
        # chain is also empty, and those should have already be optimized by
        # the search strategy
        return None
    current_delta = Delta.from_bytes(delta)
    best_base = None
    for base, data in next_deltas:
        current_delta = Delta.from_bytes(data).combine(current_delta)
        if current_delta.storage_size > max_size:
            break
        best_base = base
    return best_base



# blobstore.py - local and remote (speaking Git-LFS protocol) blob storages
#
# Copyright 2017 Facebook, Inc.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import contextlib
import errno
import hashlib
import json
import os
import re

from mercurial.i18n import _
from mercurial.node import hex

from mercurial import (
    encoding,
    error,
    httpconnection as httpconnectionmod,
    pathutil,
    pycompat,
    url as urlmod,
    util,
    vfs as vfsmod,
    worker,
)

from mercurial.utils import (
    stringutil,
    urlutil,
)

from ..largefiles import lfutil

# 64 bytes for SHA256
_lfsre = re.compile(br'\A[a-f0-9]{64}\Z')


class lfsvfs(vfsmod.vfs):
    def join(self, path: bytes | None, *insidef: bytes) -> bytes:
        """split the path at first two characters, like: XX/XXXXX..."""
        if not _lfsre.match(path):
            raise error.ProgrammingError(b'unexpected lfs path: %s' % path)
        return super().join(path[0:2], path[2:], *insidef)

    def walk(self, path=None, onerror=None):
        """Yield (dirpath, [], oids) tuple for blobs under path

        Oids only exist in the root of this vfs, so dirpath is always ''.
        """
        root = os.path.normpath(self.base)
        # when dirpath == root, dirpath[prefixlen:] becomes empty
        # because len(dirpath) < prefixlen.
        prefixlen = len(pathutil.normasprefix(root))
        oids = []

        for dirpath, dirs, files in os.walk(
            self.reljoin(self.base, path or b''), onerror=onerror
        ):
            dirpath = dirpath[prefixlen:]

            # Silently skip unexpected files and directories
            if len(dirpath) == 2:
                oids.extend(
                    [dirpath + f for f in files if _lfsre.match(dirpath + f)]
                )

        yield (b'', [], oids)


class nullvfs(lfsvfs):
    def __init__(self):
        pass

    def exists(self, path: bytes | None = None) -> bool:
        return False

    def read(self, oid):
        # store.read() calls into here if the blob doesn't exist in its
        # self.vfs.  Raise the same error as a normal vfs when asked to read a
        # file that doesn't exist.  The only difference is the full file path
        # isn't available in the error.
        raise OSError(
            errno.ENOENT,
            pycompat.sysstr(b'%s: No such file or directory' % oid),
        )

    def walk(self, path=None, onerror=None):
        return (b'', [], [])

    def write(self, *args, **kwargs) -> int:
        return 0


class lfsuploadfile(httpconnectionmod.httpsendfile):
    """a file-like object that supports keepalive."""

    def __init__(self, ui, filename):
        super().__init__(ui, filename, b'rb')
        self.read = self._data.read

    def _makeprogress(self):
        return None  # progress is handled by the worker client


class local:
    """Local blobstore for large file contents.

    This blobstore is used both as a cache and as a staging area for large blobs
    to be uploaded to the remote blobstore.
    """

    def __init__(self, repo):
        fullpath = repo.svfs.join(b'lfs/objects')
        self.vfs = lfsvfs(fullpath)

        if repo.ui.configbool(b'experimental', b'lfs.disableusercache'):
            self.cachevfs = nullvfs()
        else:
            usercache = lfutil._usercachedir(repo.ui, b'lfs')
            self.cachevfs = lfsvfs(usercache)
        self.ui = repo.ui

    def open(self, oid):
        """Open a read-only file descriptor to the named blob, in either the
        usercache or the local store."""
        return open(self.path(oid), 'rb')

    def path(self, oid):
        """Build the path for the given blob ``oid``.

        If the blob exists locally, the path may point to either the usercache
        or the local store.  If it doesn't, it will point to the local store.
        This is meant for situations where existing code that isn't LFS aware
        needs to open a blob.  Generally, prefer the ``open`` method on this
        class.
        """
        # The usercache is the most likely place to hold the file.  Commit will
        # write to both it and the local store, as will anything that downloads
        # the blobs.  However, things like clone without an update won't
        # populate the local store.  For an init + push of a local clone,
        # the usercache is the only place it _could_ be.  If not present, the
        # missing file msg here will indicate the local repo, not the usercache.
        if self.cachevfs.exists(oid):
            return self.cachevfs.join(oid)

        return self.vfs.join(oid)

    def download(self, oid, src, content_length):
        """Read the blob from the remote source in chunks, verify the content,
        and write to this local blobstore."""
        sha256 = hashlib.sha256()
        size = 0

        with self.vfs(oid, b'wb', atomictemp=True) as fp:
            for chunk in util.filechunkiter(src, size=1048576):
                fp.write(chunk)
                sha256.update(chunk)
                size += len(chunk)

            # If the server advertised a length longer than what we actually
            # received, then we should expect that the server crashed while
            # producing the response (but the server has no way of telling us
            # that), and we really don't need to try to write the response to
            # the localstore, because it's not going to match the expected.
            # The server also uses this method to store data uploaded by the
            # client, so if this happens on the server side, it's possible
            # that the client crashed or an antivirus interfered with the
            # upload.
            if content_length is not None and int(content_length) != size:
                msg = (
                    b"Response length (%d) does not match Content-Length "
                    b"header (%d) for %s"
                )
                raise LfsRemoteError(_(msg) % (size, int(content_length), oid))

            realoid = hex(sha256.digest())
            if realoid != oid:
                raise LfsCorruptionError(
                    _(b'corrupt remote lfs object: %s') % oid
                )

        self._linktousercache(oid)

    def write(self, oid, data):
        """Write blob to local blobstore.

        This should only be called from the filelog during a commit or similar.
        As such, there is no need to verify the data.  Imports from a remote
        store must use ``download()`` instead."""
        with self.vfs(oid, b'wb', atomictemp=True) as fp:
            fp.write(data)

        self._linktousercache(oid)

    def linkfromusercache(self, oid):
        """Link blobs found in the user cache into this store.

        The server module needs to do this when it lets the client know not to
        upload the blob, to ensure it is always available in this store.
        Normally this is done implicitly when the client reads or writes the
        blob, but that doesn't happen when the server tells the client that it
        already has the blob.
        """
        if not isinstance(self.cachevfs, nullvfs) and not self.vfs.exists(oid):
            self.ui.note(_(b'lfs: found %s in the usercache\n') % oid)
            lfutil.link(self.cachevfs.join(oid), self.vfs.join(oid))

    def _linktousercache(self, oid):
        # XXX: should we verify the content of the cache, and hardlink back to
        # the local store on success, but truncate, write and link on failure?
        if not self.cachevfs.exists(oid) and not isinstance(
            self.cachevfs, nullvfs
        ):
            self.ui.note(_(b'lfs: adding %s to the usercache\n') % oid)
            lfutil.link(self.vfs.join(oid), self.cachevfs.join(oid))

    def read(self, oid, verify=True):
        """Read blob from local blobstore."""
        if not self.vfs.exists(oid):
            blob = self._read(self.cachevfs, oid, verify)

            # Even if revlog will verify the content, it needs to be verified
            # now before making the hardlink to avoid propagating corrupt blobs.
            # Don't abort if corruption is detected, because `hg verify` will
            # give more useful info about the corruption- simply don't add the
            # hardlink.
            if verify or hex(hashlib.sha256(blob).digest()) == oid:
                self.ui.note(_(b'lfs: found %s in the usercache\n') % oid)
                lfutil.link(self.cachevfs.join(oid), self.vfs.join(oid))
        else:
            self.ui.note(_(b'lfs: found %s in the local lfs store\n') % oid)
            blob = self._read(self.vfs, oid, verify)
        return blob

    def _read(self, vfs, oid, verify):
        """Read blob (after verifying) from the given store"""
        blob = vfs.read(oid)
        if verify:
            _verify(oid, blob)
        return blob

    def verify(self, oid):
        """Indicate whether or not the hash of the underlying file matches its
        name."""
        sha256 = hashlib.sha256()

        with self.open(oid) as fp:
            for chunk in util.filechunkiter(fp, size=1048576):
                sha256.update(chunk)

        return oid == hex(sha256.digest())

    def has(self, oid):
        """Returns True if the local blobstore contains the requested blob,
        False otherwise."""
        return self.cachevfs.exists(oid) or self.vfs.exists(oid)


def _urlerrorreason(urlerror):
    """Create a friendly message for the given URLError to be used in an
    LfsRemoteError message.
    """
    inst = urlerror

    if isinstance(urlerror.reason, Exception):
        inst = urlerror.reason

    if hasattr(inst, 'reason'):
        try:  # usually it is in the form (errno, strerror)
            reason = inst.reason.args[1]
        except (AttributeError, IndexError):
            # it might be anything, for example a string
            reason = inst.reason
        if isinstance(reason, str):
            # SSLError of Python 2.7.9 contains a unicode
            reason = encoding.unitolocal(reason)
        return reason
    elif getattr(inst, "strerror", None):
        return encoding.strtolocal(inst.strerror)
    else:
        return stringutil.forcebytestr(urlerror)


class lfsauthhandler(util.urlreq.basehandler):
    handler_order = 480  # Before HTTPDigestAuthHandler (== 490)

    def http_error_401(self, req, fp, code, msg, headers):
        """Enforces that any authentication performed is HTTP Basic
        Authentication.  No authentication is also acceptable.
        """
        authreq = headers.get('www-authenticate', None)
        if authreq:
            scheme = authreq.split()[0]

            if scheme.lower() != 'basic':
                msg = _(b'the server must support Basic Authentication')
                raise util.urlerr.httperror(
                    req.get_full_url(),
                    code,
                    encoding.strfromlocal(msg),
                    headers,
                    fp,
                )
        return None


class _gitlfsremote:
    def __init__(self, repo, url):
        ui = repo.ui
        self.ui = ui
        baseurl, authinfo = url.authinfo()
        self.baseurl = baseurl.rstrip(b'/')
        useragent = repo.ui.config(b'experimental', b'lfs.user-agent')
        if not useragent:
            useragent = b'git-lfs/2.3.4 (Mercurial %s)' % util.version()
        self.urlopener = urlmod.opener(ui, authinfo, useragent)
        self.urlopener.add_handler(lfsauthhandler())
        self.retry = ui.configint(b'lfs', b'retry')

    def writebatch(self, pointers, fromstore):
        """Batch upload from local to remote blobstore."""
        self._batch(_deduplicate(pointers), fromstore, b'upload')

    def readbatch(self, pointers, tostore):
        """Batch download from remote to local blostore."""
        self._batch(_deduplicate(pointers), tostore, b'download')

    def _batchrequest(self, pointers, action):
        """Get metadata about objects pointed by pointers for given action

        Return decoded JSON object like {'objects': [{'oid': '', 'size': 1}]}
        See https://github.com/git-lfs/git-lfs/blob/master/docs/api/batch.md
        """
        objects = [
            {'oid': pycompat.strurl(p.oid()), 'size': p.size()}
            for p in pointers
        ]
        requestdata = pycompat.bytesurl(
            json.dumps(
                {
                    'objects': objects,
                    'operation': pycompat.strurl(action),
                }
            )
        )
        url = b'%s/objects/batch' % self.baseurl
        batchreq = util.urlreq.request(pycompat.strurl(url), data=requestdata)
        batchreq.add_header('Accept', 'application/vnd.git-lfs+json')
        batchreq.add_header('Content-Type', 'application/vnd.git-lfs+json')
        try:
            with contextlib.closing(self.urlopener.open(batchreq)) as rsp:
                rawjson = rsp.read()
        except util.urlerr.httperror as ex:
            hints = {
                400: _(
                    b'check that lfs serving is enabled on %s and "%s" is '
                    b'supported'
                )
                % (self.baseurl, action),
                404: _(b'the "lfs.url" config may be used to override %s')
                % self.baseurl,
            }
            hint = hints.get(ex.code, _(b'api=%s, action=%s') % (url, action))
            raise LfsRemoteError(
                _(b'LFS HTTP error: %s') % stringutil.forcebytestr(ex),
                hint=hint,
            )
        except util.urlerr.urlerror as ex:
            hint = (
                _(b'the "lfs.url" config may be used to override %s')
                % self.baseurl
            )
            raise LfsRemoteError(
                _(b'LFS error: %s') % _urlerrorreason(ex), hint=hint
            )
        try:
            response = pycompat.json_loads(rawjson)
        except ValueError:
            raise LfsRemoteError(
                _(b'LFS server returns invalid JSON: %s')
                % rawjson.encode("utf-8")
            )

        if self.ui.debugflag:
            self.ui.debug(b'Status: %d\n' % rsp.status)
            # lfs-test-server and hg serve return headers in different order
            headers = pycompat.bytestr(rsp.info()).strip()
            self.ui.debug(b'%s\n' % b'\n'.join(sorted(headers.splitlines())))

            if 'objects' in response:
                response['objects'] = sorted(
                    response['objects'], key=lambda p: p['oid']
                )
            self.ui.debug(
                b'%s\n'
                % pycompat.bytesurl(
                    json.dumps(
                        response,
                        indent=2,
                        separators=('', ': '),
                        sort_keys=True,
                    )
                )
            )

        def encodestr(x):
            if isinstance(x, str):
                return x.encode('utf-8')
            return x

        return pycompat.rapply(encodestr, response)

    def _checkforservererror(self, pointers, responses, action):
        """Scans errors from objects

        Raises LfsRemoteError if any objects have an error"""
        for response in responses:
            # The server should return 404 when objects cannot be found. Some
            # server implementation (ex. lfs-test-server)  does not set "error"
            # but just removes "download" from "actions". Treat that case
            # as the same as 404 error.
            if b'error' not in response:
                if action == b'download' and action not in response.get(
                    b'actions', []
                ):
                    code = 404
                else:
                    continue
            else:
                # An error dict without a code doesn't make much sense, so
                # treat as a server error.
                code = response.get(b'error').get(b'code', 500)

            ptrmap = {p.oid(): p for p in pointers}
            p = ptrmap.get(response[b'oid'], None)
            if p:
                filename = getattr(p, 'filename', b'unknown')
                errors = {
                    404: b'The object does not exist',
                    410: b'The object was removed by the owner',
                    422: b'Validation error',
                    500: b'Internal server error',
                }
                msg = errors.get(code, b'status code %d' % code)
                raise LfsRemoteError(
                    _(b'LFS server error for "%s": %s') % (filename, msg)
                )
            else:
                raise LfsRemoteError(
                    _(b'LFS server error. Unsolicited response for oid %s')
                    % response[b'oid']
                )

    def _extractobjects(self, response, pointers, action):
        """extract objects from response of the batch API

        response: parsed JSON object returned by batch API
        return response['objects'] filtered by action
        raise if any object has an error
        """
        # Scan errors from objects - fail early
        objects = response.get(b'objects', [])
        self._checkforservererror(pointers, objects, action)

        # Filter objects with given action. Practically, this skips uploading
        # objects which exist in the server.
        filteredobjects = [
            o for o in objects if action in o.get(b'actions', [])
        ]

        return filteredobjects

    def _basictransfer(self, obj, action, localstore):
        """Download or upload a single object using basic transfer protocol

        obj: dict, an object description returned by batch API
        action: string, one of ['upload', 'download']
        localstore: blobstore.local

        See https://github.com/git-lfs/git-lfs/blob/master/docs/api/\
        basic-transfers.md
        """
        oid = obj[b'oid']
        href = obj[b'actions'][action].get(b'href')
        headers = obj[b'actions'][action].get(b'header', {}).items()

        request = util.urlreq.request(pycompat.strurl(href))
        if action == b'upload':
            # If uploading blobs, read data from local blobstore.
            if not localstore.verify(oid):
                raise error.Abort(
                    _(b'detected corrupt lfs object: %s') % oid,
                    hint=_(b'run hg verify'),
                )

        for k, v in headers:
            request.add_header(pycompat.strurl(k), pycompat.strurl(v))

        try:
            if action == b'upload':
                request.data = lfsuploadfile(self.ui, localstore.path(oid))
                request.get_method = lambda: 'PUT'
                request.add_header('Content-Type', 'application/octet-stream')
                request.add_header('Content-Length', request.data.length)

            with contextlib.closing(self.urlopener.open(request)) as res:
                contentlength = res.info().get(b"content-length")
                ui = self.ui  # Shorten debug lines
                if self.ui.debugflag:
                    ui.debug(b'Status: %d\n' % res.status)
                    # lfs-test-server and hg serve return headers in different
                    # order
                    headers = pycompat.bytestr(res.info()).strip()
                    ui.debug(b'%s\n' % b'\n'.join(sorted(headers.splitlines())))

                if action == b'download':
                    # If downloading blobs, store downloaded data to local
                    # blobstore
                    localstore.download(oid, res, contentlength)
                else:
                    blocks = []
                    while True:
                        data = res.read(1048576)
                        if not data:
                            break
                        blocks.append(data)

                    response = b"".join(blocks)
                    if response:
                        ui.debug(b'lfs %s response: %s' % (action, response))
        except util.urlerr.httperror as ex:
            if self.ui.debugflag:
                self.ui.debug(
                    b'%s: %s\n' % (oid, ex.read())
                )  # XXX: also bytes?
            raise LfsRemoteError(
                _(b'LFS HTTP error: %s (oid=%s, action=%s)')
                % (stringutil.forcebytestr(ex), oid, action)
            )
        except util.urlerr.urlerror as ex:
            hint = _(b'attempted connection to %s') % pycompat.bytesurl(
                util.urllibcompat.getfullurl(request)
            )
            raise LfsRemoteError(
                _(b'LFS error: %s') % _urlerrorreason(ex), hint=hint
            )
        finally:
            if request.data:
                request.data.close()

    def _batch(self, pointers, localstore, action):
        if action not in [b'upload', b'download']:
            raise error.ProgrammingError(b'invalid Git-LFS action: %s' % action)

        response = self._batchrequest(pointers, action)
        objects = self._extractobjects(response, pointers, action)
        total = sum(x.get(b'size', 0) for x in objects)
        sizes = {}
        for obj in objects:
            sizes[obj.get(b'oid')] = obj.get(b'size', 0)
        topic = {
            b'upload': _(b'lfs uploading'),
            b'download': _(b'lfs downloading'),
        }[action]
        if len(objects) > 1:
            self.ui.note(
                _(b'lfs: need to transfer %d objects (%s)\n')
                % (len(objects), util.bytecount(total))
            )

        def transfer(chunk):
            for obj in chunk:
                objsize = obj.get(b'size', 0)
                if self.ui.verbose:
                    if action == b'download':
                        msg = _(b'lfs: downloading %s (%s)\n')
                    elif action == b'upload':
                        msg = _(b'lfs: uploading %s (%s)\n')
                    self.ui.note(
                        msg % (obj.get(b'oid'), util.bytecount(objsize))
                    )
                retry = self.retry
                while True:
                    try:
                        self._basictransfer(obj, action, localstore)
                        yield 1, obj.get(b'oid')
                        break
                    except OSError as ex:
                        if retry > 0:
                            self.ui.note(
                                _(b'lfs: failed: %r (remaining retry %d)\n')
                                % (stringutil.forcebytestr(ex), retry)
                            )
                            retry -= 1
                            continue
                        raise

        # Until https multiplexing gets sorted out.  It's not clear if
        # ConnectionManager.set_ready() is externally synchronized for thread
        # safety with Windows workers.
        if self.ui.configbool(b'experimental', b'lfs.worker-enable'):
            # The POSIX workers are forks of this process, so before spinning
            # them up, close all pooled connections.  Otherwise, there's no way
            # to coordinate between them about who is using what, and the
            # transfers will get corrupted.
            #
            # TODO: add a function to keepalive.ConnectionManager to mark all
            #  ready connections as in use, and roll that back after the fork?
            #  That would allow the existing pool of connections in this process
            #  to be preserved.
            def prefork():
                for h in self.urlopener.handlers:
                    getattr(h, "close_all", lambda: None)()

            oids = worker.worker(
                self.ui,
                0.1,
                transfer,
                (),
                sorted(objects, key=lambda o: o.get(b'oid')),
                prefork=prefork,
            )
        else:
            oids = transfer(sorted(objects, key=lambda o: o.get(b'oid')))

        with self.ui.makeprogress(
            topic, unit=_(b"bytes"), total=total
        ) as progress:
            progress.update(0)
            processed = 0
            blobs = 0
            for _one, oid in oids:
                processed += sizes[oid]
                blobs += 1
                progress.update(processed)
                self.ui.note(_(b'lfs: processed: %s\n') % oid)

        if blobs > 0:
            if action == b'upload':
                self.ui.status(
                    _(b'lfs: uploaded %d files (%s)\n')
                    % (blobs, util.bytecount(processed))
                )
            elif action == b'download':
                self.ui.status(
                    _(b'lfs: downloaded %d files (%s)\n')
                    % (blobs, util.bytecount(processed))
                )

    def __del__(self):
        # copied from mercurial/httppeer.py
        urlopener = getattr(self, 'urlopener', None)
        if urlopener:
            for h in urlopener.handlers:
                h.close()
                getattr(h, "close_all", lambda: None)()


class _dummyremote:
    """Dummy store storing blobs to temp directory."""

    def __init__(self, repo, url):
        fullpath = repo.vfs.join(b'lfs', url.path)
        self.vfs = lfsvfs(fullpath)

    def writebatch(self, pointers, fromstore):
        for p in _deduplicate(pointers):
            content = fromstore.read(p.oid(), verify=True)
            with self.vfs(p.oid(), b'wb', atomictemp=True) as fp:
                fp.write(content)

    def readbatch(self, pointers, tostore):
        for p in _deduplicate(pointers):
            with self.vfs(p.oid(), b'rb') as fp:
                tostore.download(p.oid(), fp, None)


class _nullremote:
    """Null store storing blobs to /dev/null."""

    def __init__(self, repo, url):
        pass

    def writebatch(self, pointers, fromstore):
        pass

    def readbatch(self, pointers, tostore):
        pass


class _promptremote:
    """Prompt user to set lfs.url when accessed."""

    def __init__(self, repo, url):
        pass

    def writebatch(self, pointers, fromstore, ui=None):
        self._prompt()

    def readbatch(self, pointers, tostore, ui=None):
        self._prompt()

    def _prompt(self):
        raise error.Abort(_(b'lfs.url needs to be configured'))


_storemap = {
    b'https': _gitlfsremote,
    b'http': _gitlfsremote,
    b'file': _dummyremote,
    b'null': _nullremote,
    None: _promptremote,
}


def _deduplicate(pointers):
    """Remove any duplicate oids that exist in the list"""
    reduced = util.sortdict()
    for p in pointers:
        reduced[p.oid()] = p
    return reduced.values()


def _verify(oid, content):
    realoid = hex(hashlib.sha256(content).digest())
    if realoid != oid:
        raise LfsCorruptionError(
            _(b'detected corrupt lfs object: %s') % oid,
            hint=_(b'run hg verify'),
        )


def remote(repo, remote=None):
    """remotestore factory. return a store in _storemap depending on config

    If ``lfs.url`` is specified, use that remote endpoint.  Otherwise, try to
    infer the endpoint, based on the remote repository using the same path
    adjustments as git.  As an extension, 'http' is supported as well so that
    ``hg serve`` works out of the box.

    https://github.com/git-lfs/git-lfs/blob/master/docs/api/server-discovery.md
    """
    lfsurl = repo.ui.config(b'lfs', b'url')
    url = urlutil.url(lfsurl or b'')
    if lfsurl is None:
        if remote:
            path = remote
        elif hasattr(repo, '_subtoppath'):
            # The pull command sets this during the optional update phase, which
            # tells exactly where the pull originated, whether 'paths.default'
            # or explicit.
            path = repo._subtoppath
        else:
            # TODO: investigate 'paths.remote:lfsurl' style path customization,
            # and fall back to inferring from 'paths.remote' if unspecified.
            path = repo.ui.config(b'paths', b'default') or b''

        defaulturl = urlutil.url(path)

        # TODO: support local paths as well.
        # TODO: consider the ssh -> https transformation that git applies
        if defaulturl.scheme in (b'http', b'https'):
            if defaulturl.path and defaulturl.path[:-1] != b'/':
                defaulturl.path += b'/'
            defaulturl.path = (defaulturl.path or b'') + b'.git/info/lfs'

            url = urlutil.url(bytes(defaulturl))
            repo.ui.note(_(b'lfs: assuming remote store: %s\n') % url)

    scheme = url.scheme
    if scheme not in _storemap:
        raise error.Abort(_(b'lfs: unknown url scheme: %s') % scheme)
    return _storemap[scheme](repo, url)


class LfsRemoteError(error.StorageError):
    pass


class LfsCorruptionError(error.Abort):
    """Raised when a corrupt blob is detected, aborting an operation

    It exists to allow specialized handling on the server side."""



# changelog.py - changelog class for mercurial
#
# Copyright 2005-2007 Olivia Mackall <olivia@selenic.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import typing

from .i18n import _
from .node import (
    bin,
    hex,
)
from .thirdparty import attr

# Force pytype to use the non-vendored package
if typing.TYPE_CHECKING:
    # noinspection PyPackageRequirements
    import attr

from . import (
    encoding,
    error,
    metadata,
    pycompat,
    revlog,
)
from .utils import (
    dateutil,
    stringutil,
)
from .revlogutils import (
    constants as revlog_constants,
    flagutil,
)

_defaultextra = {b'branch': b'default'}


def _string_escape(text):
    """
    >>> from .pycompat import bytechr as chr
    >>> d = {b'nl': chr(10), b'bs': chr(92), b'cr': chr(13), b'nul': chr(0)}
    >>> s = b"ab%(nl)scd%(bs)s%(bs)sn%(nul)s12ab%(cr)scd%(bs)s%(nl)s" % d
    >>> s
    'ab\\ncd\\\\\\\\n\\x0012ab\\rcd\\\\\\n'
    >>> res = _string_escape(s)
    >>> s == _string_unescape(res)
    True
    """
    # subset of the string_escape codec
    text = (
        text.replace(b'\\', b'\\\\')
        .replace(b'\n', b'\\n')
        .replace(b'\r', b'\\r')
    )
    return text.replace(b'\0', b'\\0')


def _string_unescape(text):
    if b'\\0' in text:
        # fix up \0 without getting into trouble with \\0
        text = text.replace(b'\\\\', b'\\\\\n')
        text = text.replace(b'\\0', b'\0')
        text = text.replace(b'\n', b'')
    return stringutil.unescapestr(text)


def decodeextra(text):
    """
    >>> from .pycompat import bytechr as chr
    >>> sorted(decodeextra(encodeextra({b'foo': b'bar', b'baz': chr(0) + b'2'})
    ...                    ).items())
    [('baz', '\\x002'), ('branch', 'default'), ('foo', 'bar')]
    >>> sorted(decodeextra(encodeextra({b'foo': b'bar',
    ...                                 b'baz': chr(92) + chr(0) + b'2'})
    ...                    ).items())
    [('baz', '\\\\\\x002'), ('branch', 'default'), ('foo', 'bar')]
    """
    extra = _defaultextra.copy()
    for l in text.split(b'\0'):
        if l:
            k, v = _string_unescape(l).split(b':', 1)
            extra[k] = v
    return extra


def encodeextra(d):
    # keys must be sorted to produce a deterministic changelog entry
    items = [_string_escape(b'%s:%s' % (k, d[k])) for k in sorted(d)]
    return b"\0".join(items)


def stripdesc(desc):
    """strip trailing whitespace and leading and trailing empty lines"""
    return b'\n'.join([l.rstrip() for l in desc.splitlines()]).strip(b'\n')


@attr.s
class _changelogrevision:
    # Extensions might modify _defaultextra, so let the constructor below pass
    # it in
    extra = attr.ib()
    manifest = attr.ib()
    user = attr.ib(default=b'')
    date = attr.ib(default=(0, 0))
    files = attr.ib(default=attr.Factory(list))
    filesadded = attr.ib(default=None)
    filesremoved = attr.ib(default=None)
    p1copies = attr.ib(default=None)
    p2copies = attr.ib(default=None)
    description = attr.ib(default=b'')
    branchinfo = attr.ib(default=(_defaultextra[b'branch'], False))


class changelogrevision:
    """Holds results of a parsed changelog revision.

    Changelog revisions consist of multiple pieces of data, including
    the manifest node, user, and date. This object exposes a view into
    the parsed object.
    """

    __slots__ = (
        '_offsets',
        '_text',
        '_sidedata',
        '_cpsd',
        '_changes',
    )

    def __new__(cls, cl, text, sidedata, cpsd):
        if not text:
            return _changelogrevision(extra=_defaultextra, manifest=cl.nullid)

        self = super().__new__(cls)
        # We could return here and implement the following as an __init__.
        # But doing it here is equivalent and saves an extra function call.

        # format used:
        # nodeid\n        : manifest node in ascii
        # user\n          : user, no \n or \r allowed
        # time tz extra\n : date (time is int or float, timezone is int)
        #                 : extra is metadata, encoded and separated by '\0'
        #                 : older versions ignore it
        # files\n\n       : files modified by the cset, no \n or \r allowed
        # (.*)            : comment (free text, ideally utf-8)
        #
        # changelog v0 doesn't use extra

        nl1 = text.index(b'\n')
        nl2 = text.index(b'\n', nl1 + 1)
        nl3 = text.index(b'\n', nl2 + 1)

        # The list of files may be empty. Which means nl3 is the first of the
        # double newline that precedes the description.
        if text[nl3 + 1 : nl3 + 2] == b'\n':
            doublenl = nl3
        else:
            doublenl = text.index(b'\n\n', nl3 + 1)

        self._offsets = (nl1, nl2, nl3, doublenl)
        self._text = text
        self._sidedata = sidedata
        self._cpsd = cpsd
        self._changes = None

        return self

    @property
    def manifest(self):
        return bin(self._text[0 : self._offsets[0]])

    @property
    def user(self):
        off = self._offsets
        return encoding.tolocal(self._text[off[0] + 1 : off[1]])

    @property
    def _rawdate(self):
        off = self._offsets
        dateextra = self._text[off[1] + 1 : off[2]]
        return dateextra.split(b' ', 2)[0:2]

    @property
    def _rawextra(self):
        off = self._offsets
        dateextra = self._text[off[1] + 1 : off[2]]
        fields = dateextra.split(b' ', 2)
        if len(fields) != 3:
            return None

        return fields[2]

    @property
    def date(self):
        raw = self._rawdate
        time = float(raw[0])
        # Various tools did silly things with the timezone.
        try:
            timezone = int(raw[1])
        except ValueError:
            timezone = 0

        return time, timezone

    @property
    def extra(self):
        raw = self._rawextra
        if raw is None:
            return _defaultextra

        return decodeextra(raw)

    @property
    def changes(self):
        if self._changes is not None:
            return self._changes
        if self._cpsd:
            changes = metadata.decode_files_sidedata(self._sidedata)
        else:
            changes = metadata.ChangingFiles(
                touched=self.files or (),
                added=self.filesadded or (),
                removed=self.filesremoved or (),
                p1_copies=self.p1copies or {},
                p2_copies=self.p2copies or {},
            )
        self._changes = changes
        return changes

    @property
    def files(self):
        if self._cpsd:
            return sorted(self.changes.touched)
        off = self._offsets
        if off[2] == off[3]:
            return []

        return self._text[off[2] + 1 : off[3]].split(b'\n')

    @property
    def filesadded(self):
        if self._cpsd:
            return self.changes.added
        else:
            rawindices = self.extra.get(b'filesadded')
        if rawindices is None:
            return None
        return metadata.decodefileindices(self.files, rawindices)

    @property
    def filesremoved(self):
        if self._cpsd:
            return self.changes.removed
        else:
            rawindices = self.extra.get(b'filesremoved')
        if rawindices is None:
            return None
        return metadata.decodefileindices(self.files, rawindices)

    @property
    def p1copies(self):
        if self._cpsd:
            return self.changes.copied_from_p1
        else:
            rawcopies = self.extra.get(b'p1copies')
        if rawcopies is None:
            return None
        return metadata.decodecopies(self.files, rawcopies)

    @property
    def p2copies(self):
        if self._cpsd:
            return self.changes.copied_from_p2
        else:
            rawcopies = self.extra.get(b'p2copies')
        if rawcopies is None:
            return None
        return metadata.decodecopies(self.files, rawcopies)

    @property
    def description(self):
        return encoding.tolocal(self._text[self._offsets[3] + 2 :])

    @property
    def branchinfo(self):
        extra = self.extra
        return encoding.tolocal(extra.get(b"branch")), b'close' in extra


class changelog(revlog.revlog):
    def __init__(self, opener, trypending=False, concurrencychecker=None):
        """Load a changelog revlog using an opener.

        If ``trypending`` is true, we attempt to load the index from a
        ``00changelog.i.a`` file instead of the default ``00changelog.i``.
        The ``00changelog.i.a`` file contains index (and possibly inline
        revision) data for a transaction that hasn't been finalized yet.
        It exists in a separate file to facilitate readers (such as
        hooks processes) accessing data before a transaction is finalized.

        ``concurrencychecker`` will be passed to the revlog init function, see
        the documentation there.
        """
        revlog.revlog.__init__(
            self,
            opener,
            target=(revlog_constants.KIND_CHANGELOG, None),
            radix=b'00changelog',
            checkambig=True,
            mmaplargeindex=True,
            persistentnodemap=opener.options.get(b'persistent-nodemap', False),
            concurrencychecker=concurrencychecker,
            trypending=trypending,
            may_inline=False,
        )

        if self._initempty and (self._format_version == revlog.REVLOGV1):
            # changelogs don't benefit from generaldelta.
            # (sparse-revlog has no flag, it is just a repository requirements)

            self._format_flags &= ~(
                revlog.FLAG_GENERALDELTA | revlog.FLAG_DELTA_INFO
            )
            self.delta_config.general_delta = False
            self.delta_config.sparse_revlog = False
            self.delta_config.delta_info = False
            self.data_config.generaldelta = False
            # stricly speaking we could still use sparse reading even if the
            # delta are not stored as such, however this would not yield any
            # benefit here either so we disable it for simplicity.
            self.data_config.sparse_revlog = False
            self.data_config.delta_info = False

        # Delta chains for changelogs tend to be very small because entries
        # tend to be small and don't delta well with each. So disable delta
        # chains.
        self._storedeltachains = False

        self._v2_delayed = False
        self._filteredrevs = frozenset()
        self._filteredrevs_hashcache = {}
        self._copiesstorage = opener.options.get(b'copies-storage')

    def __contains__(self, rev):
        return (0 <= rev < len(self)) and rev not in self._filteredrevs

    @property
    def filteredrevs(self):
        return self._filteredrevs

    @filteredrevs.setter
    def filteredrevs(self, val):
        # Ensure all updates go through this function
        assert isinstance(val, frozenset)
        self._filteredrevs = val
        self._filteredrevs_hashcache = {}

    def _write_docket(self, tr):
        if not self._v2_delayed:
            super()._write_docket(tr)

    def delayupdate(self, tr):
        """delay visibility of index updates to other readers"""
        assert not self._inner.is_open
        assert not self._may_inline
        # enforce that older changelog that are still inline are split at the
        # first opportunity.
        if self._inline:
            self._enforceinlinesize(tr)
        if self._docket is not None:
            self._v2_delayed = True
        else:
            new_index = self._inner.delay()
            if new_index is not None:
                self._indexfile = new_index
                tr.registertmp(new_index)
        # use "000" as prefix to make sure we run before the spliting of legacy
        # inline changelog..
        tr.addpending(b'000-cl-%i' % id(self), self._writepending)
        tr.addfinalize(b'000-cl-%i' % id(self), self._finalize)

    def _finalize(self, tr):
        """finalize index updates"""
        assert not self._inner.is_open
        if self._docket is not None:
            self._docket.write(tr)
            self._v2_delayed = False
        else:
            new_index_file = self._inner.finalize_pending()
            self._indexfile = new_index_file
            if self._inline:
                msg = 'changelog should not be inline at that point'
                raise error.ProgrammingError(msg)

    def _writepending(self, tr):
        """create a file containing the unfinalized state for
        pretxnchangegroup"""
        assert not self._inner.is_open
        if self._docket:
            any_pending = self._docket.write(tr, pending=True)
            self._v2_delayed = False
        else:
            new_index, any_pending = self._inner.write_pending()
            if new_index is not None:
                self._indexfile = new_index
                tr.registertmp(new_index)
        return any_pending

    def _enforceinlinesize(self, tr):
        if not self.is_delaying:
            revlog.revlog._enforceinlinesize(self, tr)

    def read(self, nodeorrev):
        """Obtain data from a parsed changelog revision.

        Returns a 6-tuple of:

           - manifest node in binary
           - author/user as a localstr
           - date as a 2-tuple of (time, timezone)
           - list of files
           - commit message as a localstr
           - dict of extra metadata

        Unless you need to access all fields, consider calling
        ``changelogrevision`` instead, as it is faster for partial object
        access.
        """
        d = self._revisiondata(nodeorrev)
        sidedata = self.sidedata(nodeorrev)
        copy_sd = self._copiesstorage == b'changeset-sidedata'
        c = changelogrevision(self, d, sidedata, copy_sd)
        return (c.manifest, c.user, c.date, c.files, c.description, c.extra)

    def changelogrevision(self, nodeorrev):
        """Obtain a ``changelogrevision`` for a node or revision."""
        text = self._revisiondata(nodeorrev)
        sidedata = self.sidedata(nodeorrev)
        return changelogrevision(
            self, text, sidedata, self._copiesstorage == b'changeset-sidedata'
        )

    def readfiles(self, nodeorrev):
        """
        short version of read that only returns the files modified by the cset
        """
        text = self.revision(nodeorrev)
        if not text:
            return []
        last = text.index(b"\n\n")
        l = text[:last].split(b'\n')
        return l[3:]

    def add(
        self,
        manifest,
        files,
        desc,
        transaction,
        p1,
        p2,
        user,
        date=None,
        extra=None,
    ):
        # Convert to UTF-8 encoded bytestrings as the very first
        # thing: calling any method on a localstr object will turn it
        # into a str object and the cached UTF-8 string is thus lost.
        user, desc = encoding.fromlocal(user), encoding.fromlocal(desc)

        user = user.strip()
        # An empty username or a username with a "\n" will make the
        # revision text contain two "\n\n" sequences -> corrupt
        # repository since read cannot unpack the revision.
        if not user:
            raise error.StorageError(_(b"empty username"))
        if b"\n" in user:
            raise error.StorageError(
                _(b"username %r contains a newline") % pycompat.bytestr(user)
            )

        desc = stripdesc(desc)

        if date:
            parseddate = b"%d %d" % dateutil.parsedate(date)
        else:
            parseddate = b"%d %d" % dateutil.makedate()
        if extra:
            branch = extra.get(b"branch")
            if branch in (b"default", b""):
                del extra[b"branch"]
            elif branch in (b".", b"null", b"tip"):
                raise error.StorageError(
                    _(b'the name \'%s\' is reserved') % branch
                )
        sortedfiles = sorted(files.touched)
        flags = 0
        sidedata = None
        if self._copiesstorage == b'changeset-sidedata':
            if files.has_copies_info:
                flags |= flagutil.REVIDX_HASCOPIESINFO
            sidedata = metadata.encode_files_sidedata(files)

        if extra:
            extra = encodeextra(extra)
            parseddate = b"%s %s" % (parseddate, extra)
        l = [hex(manifest), user, parseddate] + sortedfiles + [b"", desc]
        text = b"\n".join(l)
        rev = self.addrevision(
            text, transaction, len(self), p1, p2, sidedata=sidedata, flags=flags
        )
        return self.node(rev)

    def branchinfo(self, rev):
        """return the branch name and open/close state of a revision

        This function exists because creating a changectx object
        just to access this is costly."""
        return self.changelogrevision(rev).branchinfo

    def _nodeduplicatecallback(self, transaction, rev):
        # keep track of revisions that got "re-added", eg: unbunde of know rev.
        #
        # We track them in a list to preserve their order from the source bundle
        duplicates = transaction.changes.setdefault(b'revduplicates', [])
        duplicates.append(rev)



# wrapper.py - methods wrapping core mercurial logic
#
# Copyright 2017 Facebook, Inc.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import hashlib

from mercurial.i18n import _
from mercurial.node import bin, hex, short

from mercurial import (
    bundle2,
    changegroup,
    cmdutil,
    context,
    error,
    exchange,
    exthelper,
    localrepo,
    revlog,
    scmutil,
    vfs as vfsmod,
    wireprotov1server,
)

from mercurial.upgrade_utils import (
    actions as upgrade_actions,
    engine as upgrade_engine,
)

from mercurial.interfaces import repository

from mercurial.utils import (
    storageutil,
    stringutil,
)

from ..largefiles import lfutil

from . import (
    blobstore,
    pointer,
)

eh = exthelper.exthelper()


@eh.wrapfunction(localrepo, 'makefilestorage')
def localrepomakefilestorage(orig, requirements, features, **kwargs):
    if b'lfs' in requirements:
        features.add(repository.REPO_FEATURE_LFS)

    return orig(requirements=requirements, features=features, **kwargs)


@eh.wrapfunction(changegroup, 'allsupportedversions')
def allsupportedversions(orig, ui):
    versions = orig(ui)
    versions.add(b'03')
    return versions


@eh.wrapfunction(wireprotov1server, '_capabilities')
def _capabilities(orig, repo, proto):
    '''Wrap server command to announce lfs server capability'''
    caps = orig(repo, proto)
    if hasattr(repo.svfs, 'lfslocalblobstore'):
        # Advertise a slightly different capability when lfs is *required*, so
        # that the client knows it MUST load the extension.  If lfs is not
        # required on the server, there's no reason to autoload the extension
        # on the client.
        if b'lfs' in repo.requirements:
            caps.append(b'lfs-serve')

        caps.append(b'lfs')
    return caps


def bypasscheckhash(self, text):
    return False


def readfromstore(self, text):
    """Read filelog content from local blobstore transform for flagprocessor.

    Default tranform for flagprocessor, returning contents from blobstore.
    Returns a 2-typle (text, validatehash) where validatehash is True as the
    contents of the blobstore should be checked using checkhash.
    """
    p = pointer.deserialize(text)
    oid = p.oid()
    store = self.opener.lfslocalblobstore
    if not store.has(oid):
        p.filename = self.filename
        self.opener.lfsremoteblobstore.readbatch([p], store)

    # The caller will validate the content
    text = store.read(oid, verify=False)

    # pack hg filelog metadata
    hgmeta = {}
    for k in p.keys():
        if k.startswith(b'x-hg-'):
            name = k[len(b'x-hg-') :]
            hgmeta[name] = p[k]
    if hgmeta or text.startswith(b'\1\n'):
        text = storageutil.packmeta(hgmeta, text)

    return (text, True)


def writetostore(self, text):
    # hg filelog metadata (includes rename, etc)
    hgmeta, offset = storageutil.parsemeta(text)
    if offset and offset > 0:
        # lfs blob does not contain hg filelog metadata
        text = text[offset:]

    # git-lfs only supports sha256
    oid = hex(hashlib.sha256(text).digest())
    self.opener.lfslocalblobstore.write(oid, text)

    # replace contents with metadata
    longoid = b'sha256:%s' % oid
    metadata = pointer.gitlfspointer(oid=longoid, size=b'%d' % len(text))

    # by default, we expect the content to be binary. however, LFS could also
    # be used for non-binary content. add a special entry for non-binary data.
    # this will be used by filectx.isbinary().
    if not stringutil.binary(text):
        # not hg filelog metadata (affecting commit hash), no "x-hg-" prefix
        metadata[b'x-is-binary'] = b'0'

    # translate hg filelog metadata to lfs metadata with "x-hg-" prefix
    if hgmeta is not None:
        for k, v in hgmeta.items():
            metadata[b'x-hg-%s' % k] = v

    rawtext = metadata.serialize()
    return (rawtext, False)


def _islfs(rlog, node=None, rev=None):
    if rev is None:
        if node is None:
            # both None - likely working copy content where node is not ready
            return False
        rev = rlog.rev(node)
    else:
        node = rlog.node(rev)
    if node == rlog.nullid:
        return False
    flags = rlog.flags(rev)
    return bool(flags & revlog.REVIDX_EXTSTORED)


# Wrapping may also be applied by remotefilelog
def filelogaddrevision(
    orig,
    self,
    text,
    transaction,
    link,
    p1,
    p2,
    cachedelta=None,
    node=None,
    flags=revlog.REVIDX_DEFAULT_FLAGS,
    **kwds,
):
    # The matcher isn't available if reposetup() wasn't called.
    lfstrack = self._revlog.opener.options.get(b'lfstrack')

    if lfstrack:
        textlen = len(text)
        # exclude hg rename meta from file size
        meta, offset = storageutil.parsemeta(text)
        if offset:
            textlen -= offset

        if lfstrack(self._revlog.filename, textlen):
            flags |= revlog.REVIDX_EXTSTORED

    return orig(
        self,
        text,
        transaction,
        link,
        p1,
        p2,
        cachedelta=cachedelta,
        node=node,
        flags=flags,
        **kwds,
    )


# Wrapping may also be applied by remotefilelog
def filelogrenamed(orig, self, node):
    if _islfs(self._revlog, node):
        rawtext = self._revlog.rawdata(node)
        if not rawtext:
            return False
        metadata = pointer.deserialize(rawtext)
        if b'x-hg-copy' in metadata and b'x-hg-copyrev' in metadata:
            return metadata[b'x-hg-copy'], bin(metadata[b'x-hg-copyrev'])
        else:
            return False
    return orig(self, node)


# Wrapping may also be applied by remotefilelog
def filelog_has_meta(orig, self, node):
    if _islfs(self._revlog, node):
        rawtext = self._revlog.rawdata(node)
        if not rawtext:
            return False
        metadata = pointer.deserialize(rawtext)
        for k in metadata:
            if k.startswith(b'x-hg-'):
                return True
        return False
    return orig(self, node)


# Wrapping may also be applied by remotefilelog
def filelogsize(orig, self, rev):
    if _islfs(self._revlog, rev=rev):
        # fast path: use lfs metadata to answer size
        rawtext = self._revlog.rawdata(rev)
        metadata = pointer.deserialize(rawtext)
        return int(metadata[b'size'])
    return orig(self, rev)


@eh.wrapfunction(revlog, '_verify_revision')
def _verify_revision(orig, rl, skipflags, state, node):
    if _islfs(rl, node=node):
        rawtext = rl.rawdata(node)
        metadata = pointer.deserialize(rawtext)

        # Don't skip blobs that are stored locally, as local verification is
        # relatively cheap and there's no other way to verify the raw data in
        # the revlog.
        if rl.opener.lfslocalblobstore.has(metadata.oid()):
            skipflags &= ~revlog.REVIDX_EXTSTORED
        elif skipflags & revlog.REVIDX_EXTSTORED:
            # The wrapped method will set `skipread`, but there's enough local
            # info to check renames.
            state[b'safe_renamed'].add(node)

    orig(rl, skipflags, state, node)


@eh.wrapfunction(context.basefilectx, 'cmp')
def filectxcmp(orig, self, fctx):
    """returns True if text is different than fctx"""
    # some fctx (ex. hg-git) is not based on basefilectx and do not have islfs
    if self.islfs() and getattr(fctx, 'islfs', lambda: False)():
        # fast path: check LFS oid
        p1 = pointer.deserialize(self.rawdata())
        p2 = pointer.deserialize(fctx.rawdata())
        return p1.oid() != p2.oid()
    return orig(self, fctx)


@eh.wrapfunction(context.basefilectx, 'isbinary')
def filectxisbinary(orig, self):
    if self.islfs():
        # fast path: use lfs metadata to answer isbinary
        metadata = pointer.deserialize(self.rawdata())
        # if lfs metadata says nothing, assume it's binary by default
        return bool(int(metadata.get(b'x-is-binary', 1)))
    return orig(self)


def filectxislfs(self):
    return _islfs(self.filelog()._revlog, self.filenode())


@eh.wrapfunction(cmdutil, '_updatecatformatter')
def _updatecatformatter(orig, fm, ctx, matcher, path, decode):
    orig(fm, ctx, matcher, path, decode)
    fm.data(rawdata=ctx[path].rawdata())


@eh.wrapfunction(scmutil, 'wrapconvertsink')
def convertsink(orig, sink):
    sink = orig(sink)
    if sink.repotype == b'hg':

        class lfssink(sink.__class__):
            def putcommit(
                self,
                files,
                copies,
                parents,
                commit,
                source,
                revmap,
                full,
                cleanp2,
            ):
                pc = super().putcommit
                node = pc(
                    files,
                    copies,
                    parents,
                    commit,
                    source,
                    revmap,
                    full,
                    cleanp2,
                )

                if b'lfs' not in self.repo.requirements:
                    ctx = self.repo[node]

                    # The file list may contain removed files, so check for
                    # membership before assuming it is in the context.
                    if any(f in ctx and ctx[f].islfs() for f, n in files):
                        self.repo.requirements.add(b'lfs')
                        scmutil.writereporequirements(self.repo)

                return node

        sink.__class__ = lfssink

    return sink


# bundlerepo uses "vfsmod.readonlyvfs(othervfs)", we need to make sure lfs
# options and blob stores are passed from othervfs to the new readonlyvfs.
@eh.wrapfunction(vfsmod.readonlyvfs, '__init__')
def vfsinit(orig, self, othervfs):
    orig(self, othervfs)
    # copy lfs related options
    for k, v in othervfs.options.items():
        if k.startswith(b'lfs'):
            self.options[k] = v
    # also copy lfs blobstores. note: this can run before reposetup, so lfs
    # blobstore attributes are not always ready at this time.
    for name in ['lfslocalblobstore', 'lfsremoteblobstore']:
        if hasattr(othervfs, name):
            setattr(self, name, getattr(othervfs, name))


def _prefetchfiles(repo, revmatches):
    """Ensure that required LFS blobs are present, fetching them as a group if
    needed."""
    if not hasattr(repo.svfs, 'lfslocalblobstore'):
        return

    pointers = []
    oids = set()
    localstore = repo.svfs.lfslocalblobstore

    for rev, match in revmatches:
        ctx = repo[rev]
        for f in ctx.walk(match):
            p = pointerfromctx(ctx, f)
            if p and p.oid() not in oids and not localstore.has(p.oid()):
                p.filename = f
                pointers.append(p)
                oids.add(p.oid())

    if pointers:
        # Recalculating the repo store here allows 'paths.default' that is set
        # on the repo by a clone command to be used for the update.
        blobstore.remote(repo).readbatch(pointers, localstore)


def _canskipupload(repo):
    # Skip if this hasn't been passed to reposetup()
    if not hasattr(repo.svfs, 'lfsremoteblobstore'):
        return True

    # if remotestore is a null store, upload is a no-op and can be skipped
    return isinstance(repo.svfs.lfsremoteblobstore, blobstore._nullremote)


def candownload(repo):
    # Skip if this hasn't been passed to reposetup()
    if not hasattr(repo.svfs, 'lfsremoteblobstore'):
        return False

    # if remotestore is a null store, downloads will lead to nothing
    return not isinstance(repo.svfs.lfsremoteblobstore, blobstore._nullremote)


def uploadblobsfromrevs(repo, revs):
    """upload lfs blobs introduced by revs"""
    if _canskipupload(repo):
        return
    pointers = extractpointers(repo, revs)
    uploadblobs(repo, pointers)


def prepush(pushop):
    """Prepush hook.

    Read through the revisions to push, looking for filelog entries that can be
    deserialized into metadata so that we can block the push on their upload to
    the remote blobstore.
    """
    return uploadblobsfromrevs(pushop.repo, pushop.outgoing.missing)


@eh.wrapfunction(exchange, 'push')
def push(orig, repo, remote, *args, **kwargs):
    """bail on push if the extension isn't enabled on remote when needed, and
    update the remote store based on the destination path."""
    if b'lfs' in repo.requirements:
        # If the remote peer is for a local repo, the requirement tests in the
        # base class method enforce lfs support.  Otherwise, some revisions in
        # this repo use lfs, and the remote repo needs the extension loaded.
        if not remote.local() and not remote.capable(b'lfs'):
            # This is a copy of the message in exchange.push() when requirements
            # are missing between local repos.
            m = _(b"required features are not supported in the destination: %s")
            raise error.Abort(
                m % b'lfs', hint=_(b'enable the lfs extension on the server')
            )

        # Repositories where this extension is disabled won't have the field.
        # But if there's a requirement, then the extension must be loaded AND
        # there may be blobs to push.
        remotestore = repo.svfs.lfsremoteblobstore
        try:
            repo.svfs.lfsremoteblobstore = blobstore.remote(repo, remote.url())
            return orig(repo, remote, *args, **kwargs)
        finally:
            repo.svfs.lfsremoteblobstore = remotestore
    else:
        return orig(repo, remote, *args, **kwargs)


# when writing a bundle via "hg bundle" command, upload related LFS blobs
@eh.wrapfunction(bundle2, 'writenewbundle')
def writenewbundle(
    orig, ui, repo, source, filename, bundletype, outgoing, *args, **kwargs
):
    """upload LFS blobs added by outgoing revisions on 'hg bundle'"""
    uploadblobsfromrevs(repo, outgoing.missing)
    return orig(
        ui, repo, source, filename, bundletype, outgoing, *args, **kwargs
    )


def extractpointers(repo, revs):
    """return a list of lfs pointers added by given revs"""
    repo.ui.debug(b'lfs: computing set of blobs to upload\n')
    pointers = {}

    makeprogress = repo.ui.makeprogress
    with makeprogress(
        _(b'lfs search'), _(b'changesets'), len(revs)
    ) as progress:
        for r in revs:
            ctx = repo[r]
            for p in pointersfromctx(ctx).values():
                pointers[p.oid()] = p
            progress.increment()
        return sorted(pointers.values(), key=lambda p: p.oid())


def pointerfromctx(ctx, f, removed=False):
    """return a pointer for the named file from the given changectx, or None if
    the file isn't LFS.

    Optionally, the pointer for a file deleted from the context can be returned.
    Since no such pointer is actually stored, and to distinguish from a non LFS
    file, this pointer is represented by an empty dict.
    """
    _ctx = ctx
    if f not in ctx:
        if not removed:
            return None
        if f in ctx.p1():
            _ctx = ctx.p1()
        elif f in ctx.p2():
            _ctx = ctx.p2()
        else:
            return None
    fctx = _ctx[f]
    if not _islfs(fctx.filelog()._revlog, fctx.filenode()):
        return None
    try:
        p = pointer.deserialize(fctx.rawdata())
        if ctx == _ctx:
            return p
        return {}
    except pointer.InvalidPointer as ex:
        raise error.Abort(
            _(b'lfs: corrupted pointer (%s@%s): %s\n')
            % (f, short(_ctx.node()), ex)
        )


def pointersfromctx(ctx, removed=False):
    """return a dict {path: pointer} for given single changectx.

    If ``removed`` == True and the LFS file was removed from ``ctx``, the value
    stored for the path is an empty dict.
    """
    result = {}
    m = ctx.repo().narrowmatch()

    # TODO: consider manifest.fastread() instead
    for f in ctx.files():
        if not m(f):
            continue
        p = pointerfromctx(ctx, f, removed=removed)
        if p is not None:
            result[f] = p
    return result


def uploadblobs(repo, pointers):
    """upload given pointers from local blobstore"""
    if not pointers:
        return

    remoteblob = repo.svfs.lfsremoteblobstore
    remoteblob.writebatch(pointers, repo.svfs.lfslocalblobstore)


@eh.wrapfunction(upgrade_engine, 'finishdatamigration')
def upgradefinishdatamigration(orig, ui, srcrepo, dstrepo, requirements):
    orig(ui, srcrepo, dstrepo, requirements)

    # Skip if this hasn't been passed to reposetup()
    if hasattr(srcrepo.svfs, 'lfslocalblobstore') and hasattr(
        dstrepo.svfs, 'lfslocalblobstore'
    ):
        srclfsvfs = srcrepo.svfs.lfslocalblobstore.vfs
        dstlfsvfs = dstrepo.svfs.lfslocalblobstore.vfs

        for dirpath, dirs, files in srclfsvfs.walk():
            for oid in files:
                ui.write(_(b'copying lfs blob %s\n') % oid)
                lfutil.link(srclfsvfs.join(oid), dstlfsvfs.join(oid))


@eh.wrapfunction(upgrade_actions, 'preservedrequirements')
@eh.wrapfunction(upgrade_actions, 'supporteddestrequirements')
def upgraderequirements(orig, repo):
    reqs = orig(repo)
    if b'lfs' in repo.requirements:
        reqs.add(b'lfs')
    return reqs



# demandimportpy3 - global demand-loading of modules for Mercurial
#
# Copyright 2017 Facebook Inc.
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

"""Lazy loading for Python 3.6 and above.

This uses the new importlib finder/loader functionality available in Python 3.5
and up. The code reuses most of the mechanics implemented inside importlib.util,
but with a few additions:

* Allow excluding certain modules from lazy imports.
* Expose an interface that's substantially the same as demandimport for
  Python 2.

This also has some limitations compared to the Python 2 implementation:

* Much of the logic is per-package, not per-module, so any packages loaded
  before demandimport is enabled will not be lazily imported in the future. In
  practice, we only expect builtins to be loaded before demandimport is
  enabled.
"""

from __future__ import annotations

import contextlib
import importlib.util
import sys

from . import tracing

_deactivated = False


class _lazyloaderex(importlib.util.LazyLoader):
    """This is a LazyLoader except it also follows the _deactivated global and
    the ignore list.
    """

    _HAS_DYNAMIC_ATTRIBUTES = True  # help pytype not flag self.loader

    def exec_module(self, module):
        """Make the module load lazily."""
        with tracing.log('demandimport %s', module):
            if _deactivated or module.__name__ in ignores:
                # Reset the loader on the module as super() does (issue6725)
                module.__spec__.loader = self.loader
                module.__loader__ = self.loader

                self.loader.exec_module(module)
            else:
                super().exec_module(module)


class LazyFinder:
    """A wrapper around a ``MetaPathFinder`` that makes loaders lazy.

    ``sys.meta_path`` finders have their ``find_spec()`` called to locate a
    module. This returns a ``ModuleSpec`` if found or ``None``. The
    ``ModuleSpec`` has a ``loader`` attribute, which is called to actually
    load a module.

    Our class wraps an existing finder and overloads its ``find_spec()`` to
    replace the ``loader`` with our lazy loader proxy.

    We have to use __getattribute__ to proxy the instance because some meta
    path finders don't support monkeypatching.
    """

    __slots__ = ("_finder",)

    def __init__(self, finder):
        object.__setattr__(self, "_finder", finder)

    def __repr__(self):
        return "<LazyFinder for %r>" % object.__getattribute__(self, "_finder")

    # __bool__ is canonical Python 3. But check-code insists on __nonzero__ being
    # defined via `def`.
    def __nonzero__(self):
        return bool(object.__getattribute__(self, "_finder"))

    __bool__ = __nonzero__

    def __getattribute__(self, name):
        if name in ("_finder", "find_spec"):
            return object.__getattribute__(self, name)

        return getattr(object.__getattribute__(self, "_finder"), name)

    def __delattr__(self, name):
        return delattr(object.__getattribute__(self, "_finder"), name)

    def __setattr__(self, name, value):
        return setattr(object.__getattribute__(self, "_finder"), name, value)

    def find_spec(self, fullname, path, target=None):
        finder = object.__getattribute__(self, "_finder")
        try:
            find_spec = finder.find_spec
        except AttributeError:
            loader = finder.find_module(fullname, path)
            if loader is None:
                spec = None
            else:
                spec = importlib.util.spec_from_loader(fullname, loader)
        else:
            spec = find_spec(fullname, path, target)

        # Lazy loader requires exec_module().
        if (
            spec is not None
            and spec.loader is not None
            and getattr(spec.loader, "exec_module", None)
        ):
            spec.loader = _lazyloaderex(spec.loader)

        return spec


ignores = set()


def init(ignoreset):
    global ignores
    ignores = ignoreset


def isenabled():
    return not _deactivated and any(
        isinstance(finder, LazyFinder) for finder in sys.meta_path
    )


def disable():
    new_finders = []
    for finder in sys.meta_path:
        new_finders.append(
            finder._finder if isinstance(finder, LazyFinder) else finder
        )
    sys.meta_path[:] = new_finders


def enable():
    new_finders = []
    for finder in sys.meta_path:
        new_finders.append(
            LazyFinder(finder) if not isinstance(finder, LazyFinder) else finder
        )
    sys.meta_path[:] = new_finders


@contextlib.contextmanager
def deactivated():
    # This implementation is a bit different from Python 2's. Python 3
    # maintains a per-package finder cache in sys.path_importer_cache (see
    # PEP 302). This means that we can't just call disable + enable.
    # If we do that, in situations like:
    #
    #   demandimport.enable()
    #   ...
    #   from foo.bar import mod1
    #   with demandimport.deactivated():
    #       from foo.bar import mod2
    #
    # mod2 will be imported lazily. (The converse also holds -- whatever finder
    # first gets cached will be used.)
    #
    # Instead, have a global flag the LazyLoader can use.
    global _deactivated
    demandenabled = isenabled()
    if demandenabled:
        _deactivated = True
    try:
        yield
    finally:
        if demandenabled:
            _deactivated = False



# policy.py - module policy logic for Mercurial.
#
# Copyright 2015 Gregory Szorc <gregory.szorc@gmail.com>
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.

from __future__ import annotations

import os
import sys


# Rules for how modules can be loaded. Values are:
#
#    c - require C extensions
#    rust+c - require Rust and C extensions
#    rust+c-allow - allow Rust and C extensions with fallback to pure Python
#                   for each
#    allow - allow pure Python implementation when C loading fails
#    cffi - required cffi versions (implemented within pure module)
#    cffi-allow - allow pure Python implementation if cffi version is missing
#    py - only load pure Python modules
#
# By default, fall back to the pure modules so the in-place build can
# run without recompiling the C extensions. This will be overridden by
# __modulepolicy__ generated by setup.py.
policy: bytes = b'allow'
_packageprefs: dict[bytes, tuple[str | None, str | None]] = {
    # policy: (versioned package, pure package)
    b'c': ('cext', None),
    b'allow': ('cext', 'pure'),
    b'cffi': ('cffi', None),
    b'cffi-allow': ('cffi', 'pure'),
    b'py': (None, 'pure'),
    # For now, rust policies impact importrust only
    b'rust+c': ('cext', None),
    b'rust+c-allow': ('cext', 'pure'),
}

try:
    from . import __modulepolicy__  # pytype: disable=import-error

    policy: bytes = __modulepolicy__.modulepolicy
except ImportError:
    pass

# PyPy doesn't load C extensions.
#
# The canonical way to do this is to test platform.python_implementation().
# But we don't import platform and don't bloat for it here.
if '__pypy__' in sys.builtin_module_names:
    policy: bytes = b'cffi'

# Environment variable can always force settings.
if os.environ.get('HGMODULEPOLICY'):  # ignore None and Empty
    policy: bytes = os.environ['HGMODULEPOLICY'].encode('utf-8')


def _importfrom(pkgname: str, modname: str):
    # from .<pkgname> import <modname> (where . is looked through this module)
    fakelocals = {}
    pkg = __import__(pkgname, globals(), fakelocals, [modname], level=1)
    try:
        fakelocals[modname] = mod = getattr(pkg, modname)
    except AttributeError:
        raise ImportError('cannot import name %s' % modname)
    # force import; fakelocals[modname] may be replaced with the real module
    getattr(mod, '__doc__', None)
    return fakelocals[modname]


# keep in sync with "version" in C modules
_cextversions: dict[tuple[str, str], int] = {
    ('cext', 'base85'): 1,
    ('cext', 'bdiff'): 3,
    ('cext', 'mpatch'): 1,
    ('cext', 'osutil'): 4,
    ('cext', 'parsers'): 21,
}

# map import request to other package or module
_modredirects: dict[tuple[str, str], tuple[str, str]] = {
    ('cext', 'charencode'): ('cext', 'parsers'),
    ('cffi', 'base85'): ('pure', 'base85'),
    ('cffi', 'charencode'): ('pure', 'charencode'),
    ('cffi', 'parsers'): ('pure', 'parsers'),
}


def _checkmod(pkgname: str, modname: str, mod) -> None:
    expected = _cextversions.get((pkgname, modname))
    actual = getattr(mod, 'version', None)
    if actual != expected:
        raise ImportError(
            'cannot import module %s.%s '
            '(expected version: %d, actual: %r)'
            % (pkgname, modname, expected, actual)
        )


def importmod(modname: str):
    """Import module according to policy and check API version"""
    try:
        verpkg, purepkg = _packageprefs[policy]
    except KeyError:
        raise ImportError('invalid HGMODULEPOLICY %r' % policy)
    assert verpkg or purepkg
    if verpkg:
        pn, mn = _modredirects.get((verpkg, modname), (verpkg, modname))
        try:
            mod = _importfrom(pn, mn)
            if pn == verpkg:
                _checkmod(pn, mn, mod)
            return mod
        except ImportError:
            if not purepkg:
                raise
    pn, mn = _modredirects.get((purepkg, modname), (purepkg, modname))
    return _importfrom(pn, mn)


def _isrustpermissive() -> bool:
    """Assuming the policy is a Rust one, tell if it's permissive."""
    return policy.endswith(b'-allow')


def has_rust() -> bool:
    """Returns whether the policy contains Rust in any fashion"""
    return b"rust" in policy


def importrust(modname: str, member: str | None = None, default=None):
    """Import Rust module according to policy and availability.

    If policy isn't a Rust one, this returns `default`.

    If either the module or its member is not available, this returns `default`
    if policy is permissive and raises `ImportError` if not.
    """
    if not policy.startswith(b'rust'):
        return default

    dlib_name = "pyo3_rustext"

    try:
        mod = _importfrom(dlib_name, modname)
    except ImportError:
        if _isrustpermissive():
            return default
        raise
    if member is None:
        return mod

    try:
        return getattr(mod, member)
    except AttributeError:
        if _isrustpermissive():
            return default
        raise ImportError("Cannot import name %s" % member)



# mercurial/interfaces/_basetypes.py - internal base type aliases for interfaces
#
# This software may be used and distributed according to the terms of the
# GNU General Public License version 2 or any later version.
#
# This module contains trivial type aliases that other interfaces might need,
# in a location designed to avoid import cycles.  This is for internal usage
# by the modules in `mercurial.interfaces`, instead of importing the `types`
# module.
#
# For using type aliases outside `mercurial.interfaces`, look at the
# `mercurial.interfaces.types` module.

from __future__ import annotations

from typing import Any

UserMsgT = bytes
"""Text (maybe) displayed to the user."""

HgPathT = bytes
"""A path usable with Mercurial's vfs."""

FsPathT = bytes
"""A path on disk (after vfs encoding)."""

# I doubt we will be able to any anything more precise cheaply, however this is
# useful to richer annotations.
NodeIdT = bytes
"""a nodeid identifier"""

# TODO: create a Protocol class,
RepoT = Any

# TODO: create a Protocol class,
RevlogT = Any

# I doubt we will be able to any anything more precise cheaply, however this is
# useful to richer annotations.
RevnumT = int
"""a revision number"""

# See revset.matchany()
RevsetAliasesT = dict[bytes, bytes]
"""A mapping of a revset alias name to its (string) definition."""

# TODO: create a Protocol class,
UiT = Any

# TODO: make a protocol class for this
VfsT = Any

VfsKeyT = bytes
"""Vfs identifier, typically used in a VfsMap."""

CallbackCategoryT = bytes
"""Key identifying a callback category."""
